# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from collections.abc import Collection
from random import randint
from typing import Any
from copy import deepcopy

from airflow.sdk import Context
from airflow.sdk.types import RuntimeTaskInstanceProtocol

from airflow.providers.nomad.exceptions import NomadTaskOperatorError
from airflow.providers.nomad.generic_interfaces.nomad_operator_interface import NomadOperator
from airflow.providers.nomad.templates.job_template import DEFAULT_TASK_TEMPLATE_PYTHON
from airflow.providers.nomad.utils import job_id_from_taskinstance


class NomadPythonTaskOperator(NomadOperator):
    """Nomad Operator allowing for lightweight job submission"""

    template_fields: Collection[str] = [
        "template_path",
        "template_content",
        "env",
        "image",
        "entrypoint",
        "args",
        "command",
        "task_resources",
        "ephemeral_disk",
        "volumes",
        "volume_mounts",
        "python_command",
    ]

    def __init__(
        self,
        python_command: str,
        template_path: str | None = None,
        template_content: str | None = None,
        env: dict[str, str] | None = None,
        image: str | None = None,
        entrypoint: list[str] | None = None,
        args: list[str] | None = None,
        command: str | None = None,
        task_resources: dict[str, Any] | None = None,
        ephemeral_disk: dict[str, str] | None = None,
        volumes: dict[str, dict[str, int | bool | str]] | None = None,
        volume_mounts: list[dict[str, int | bool | str]] | None = None,
        **kwargs,
    ):
        if template_path and template_content:
            raise ValueError("Only one of 'template_content' and 'template_path' can be specified")
        self.template_content = template_content
        self.template_path = template_path
        self.image = image
        self.entrypoint = entrypoint
        self.args = args
        self.command = command
        self.env = env
        self.task_resources = task_resources
        self.ephemeral_disk = ephemeral_disk
        self.volumes = volumes
        self.volume_mounts = volume_mounts
        self.python_command = python_command
        super().__init__(observe=True, **kwargs)

    def job_id(self, ti: RuntimeTaskInstanceProtocol):
        id_base = job_id_from_taskinstance(ti)
        rnd = randint(0, 10000)
        while self.nomad_mgr.get_nomad_job_submission(f"{id_base}-{rnd}"):
            rnd = randint(0, 10000)
        return f"{id_base}-{rnd}"

    def param_defined(self, param: str, context: Context) -> Any | None:
        if value := getattr(self, param, None):
            return value
        return context.get("params", {}).get(param)

    @property
    def tpl_attrs_dict(self):
        attrs = [
            "template_path",
            "template_content",
            "image",
            "entrypoint",
            "args",
            "command",
            "env",
            "task_resources",
            "volumes",
            "volume_mounts",
            "ephemeral_disk",
        ]
        return {attr: getattr(self, attr) for attr in attrs}

    def prepare_job_template(self, context: Context):
        updated_context = deepcopy(context.get("params", {}))
        self_attrs = {
            attr: self.tpl_attrs_dict[attr]
            for attr in self.tpl_attrs_dict
            if self.tpl_attrs_dict[attr] is not None
        }
        updated_context.update(self_attrs)

        if not (
            template := self.nomad_mgr.prepare_job_template(
                updated_context, DEFAULT_TASK_TEMPLATE_PYTHON
            )
        ):
            raise NomadTaskOperatorError(f"No template for task with context {context}")

        if not (ti := context.get("ti")):
            raise NomadTaskOperatorError(f"No task instance found in context {context}")

        if template.Job.TaskGroups[0].Tasks[0].Config.args:
            template.Job.TaskGroups[0].Tasks[0].Config.args.append(self.python_command)
        else:
            template.Job.TaskGroups[0].Tasks[0].Config.args = [self.python_command]

        template.Job.ID = self.job_id(ti)
        self.template = template



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Nomad Executor.
"""

from __future__ import annotations

import argparse
import logging
from typing import Any

from airflow.configuration import conf
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.utils.log.file_task_handler import FileTaskHandler
from airflow.utils.log.logging_mixin import remove_escape_codes
from airflow.utils.state import TaskInstanceState
from pydantic import ValidationError

from airflow.providers.nomad.constants import CONFIG_SECTION
from airflow.providers.nomad.exceptions import NomadProviderException
from airflow.providers.nomad.generic_interfaces.executor_interface import ExecutorInterface
from airflow.providers.nomad.log import NomadLogHandler
from airflow.providers.nomad.manager import NomadManager
from airflow.providers.nomad.models import NomadJobModel
from airflow.providers.nomad.templates.job_template import (
    DEFAULT_JOB_NAME,
    DEFAULT_TASK_TEMPLATE_SDK,
    SDK_ENTRYPOINT,
)
from airflow.providers.nomad.utils import (
    job_id_from_taskinstance_key,
    job_task_id_from_taskinstance_key,
)

logger = logging.getLogger(__name__)


class NomadExecutor(ExecutorInterface):
    """Executor for Nomad."""

    RUNNING_POD_LOG_LINES = 100
    supports_ad_hoc_ti_run: bool = True

    def __init__(self):
        self.parallelism: int = conf.getint(CONFIG_SECTION, "parallelism", fallback=128)
        self.nomad_mgr = NomadManager()
        super().__init__(parallelism=self.parallelism)

    def start(self) -> None:
        """Start the executor."""
        self.log.info("Starting Nomad executor")
        self.nomad_mgr.initialize()

    def validate_exeucutor_config(self, executor_config: dict | None) -> dict | None:
        if not executor_config:
            return executor_config

        remove = ["args"]
        warn = ["entrypoint", "command"]
        for field in remove:
            if executor_config.pop(field, None):
                self.log.error(
                    f"'{field}' is an invalid parameter for the executor, will be ignored"
                )

        for field in warn:
            if field in executor_config:
                self.log.warning(
                    f"'{field}' should be used in a way that '{SDK_ENTRYPOINT} + <actual command>' will be used as 'args'"
                )

        return executor_config

    def prepare_job_template(
        self, key: TaskInstanceKey, command: list[str], executor_config: dict | None = None
    ) -> dict[str, Any]:
        """Adjutst template to suit upcoming job execution

        :param key: reference to the task instance in question
        :return: job template as as dictionary
        """
        job_id = job_id_from_taskinstance_key(key)
        job_task_id = job_task_id_from_taskinstance_key(key)

        valid_config = self.validate_exeucutor_config(executor_config)
        if not (
            job_model := self.nomad_mgr.prepare_job_template(
                valid_config, default_template=DEFAULT_TASK_TEMPLATE_SDK
            )
        ):
            raise NomadProviderException("Couldn't retrieve job template")

        if job_model.Job.Name == DEFAULT_JOB_NAME:
            job_model.Job.Name = f"airflow-run-{job_task_id}-{key[3]}"

        if job_model.Job.TaskGroups[0].Tasks[0].Config.entrypoint != SDK_ENTRYPOINT:
            job_model.Job.TaskGroups[0].Tasks[0].Config.args = SDK_ENTRYPOINT + command
        else:
            job_model.Job.TaskGroups[0].Tasks[0].Config.args = command

        job_model.Job.TaskGroups[0].Tasks[0].Name = job_task_id
        job_model.Job.ID = job_id

        self.log.debug(
            f"Command running: python -m airflow.sdk.execution_time.execute_workload --json-string '{command[0]}'\n"
        )
        return job_model.model_dump(exclude_unset=True)

    def run_job(self, job_template: dict[str, Any] | None) -> str | None:
        """Execute the job defined by a potential job template

        :param: Job template corresponding to the job
        :return: No news is good news, or the error that occured on execution attempt
        """
        if not job_template:
            raise NomadProviderException("Job template missing")

        try:
            job_model = NomadJobModel.model_validate(job_template)
        except ValidationError:
            raise NomadProviderException("Job template doesn't comply to expected format ({err})")

        return self.nomad_mgr.register_job(job_model)

    def remove_job_if_hanging(self, key: TaskInstanceKey) -> tuple[TaskInstanceState, str] | None:
        """Whether the job failed outside of the Airflow context

        :param key: reference to the task instance in question
        :return: either a tuple of: (task status to set (typically: FAILED), additional info)
                 or None if no data could be retrieved for the job
        """
        job_id = job_id_from_taskinstance_key(key)
        if not (outcome := self.nomad_mgr.remove_job_if_hanging(job_id)):
            return None

        killed, error_msg = outcome
        if killed:
            return TaskInstanceState.FAILED, error_msg
        return None

    def _get_task_log(
        self, ti: TaskInstance, try_number: int, stderr=False
    ) -> tuple[list[str], list[str]]:
        messages = []
        log = []
        logtype = "error" if stderr else "standard"
        try:
            messages.append(
                f"Attempting to fetch {logtype} logs for task {ti.key} through Nomad API (attempts: {try_number})"
            )
            messages_received, logs_received = self.retrieve_logs(ti.key, stderr=stderr)
            messages += messages_received

            for line in logs_received:
                log.append(remove_escape_codes(line))
            if log:
                messages.append(f"Found {logtype} logs for running job via Nomad API")
        except Exception as e:
            messages.append(f"Reading {logtype} logs failed: {e}")

        # In case the task didn't even make it to be submitted, we may be able to get info about reasons from Nomad
        if not log and not stderr:
            job_id = job_id_from_taskinstance_key(ti.key)
            if job_info := self.nomad_mgr.job_all_info_str(job_id):
                messages.append("Nomad job evaluations retrieved")
                log.append(
                    "No task logs found, but the following information was retrieved from Nomad:"
                )
                log += job_info

        return messages, log

    def get_task_log(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        """Universal way to retreive logs

        Retrieving logs so that it's compatible both with FileTaskHandler, and
        leaves space for NomadLogHandler grouping feature.

        :params ti: TaskInstance in question
        :return: a list of operational messages together with the logs (as a list of entries)
        """
        messages, logs = self._get_task_log(ti, try_number)
        if (
            conf.get("logging", "task_log_reader", fallback=FileTaskHandler.name)
            != NomadLogHandler.name
        ):
            messages_err, logs_err = self._get_task_log(ti, try_number, stderr=True)
            if logs_err:
                logs = logs + logs_err
                messages = messages + messages_err
        return messages, logs

    def get_task_stderr(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        return self._get_task_log(ti, try_number, stderr=True)

    def retrieve_logs(self, key: TaskInstanceKey, stderr=False) -> tuple[list[str], list[str]]:
        # Note: this method is executed by the FileTaskHandler and not the Scheduler
        # We have no access to the running NomadExecutor's state
        if not self.nomad_mgr.nomad:
            self.start()

        messages = []
        logs = ""
        job_id = job_id_from_taskinstance_key(key)
        allocations = self.nomad_mgr.get_nomad_job_allocation(job_id)

        if not isinstance(allocations, list):
            messages.append("Unexpected result from Nomad API allocations query")
        elif len(allocations) == 0:
            messages.append(f"No allocations found for {job_id}")
        else:
            for allocation in allocations:
                if len(allocations) > 1:
                    logs += f"\nAllocation ID {allocation.ID}:\n"

                # Normally this is single element, as the executor insists on a single task
                for task_name in allocation.TaskStates:
                    if len(allocation.TaskStates) > 1:
                        logs += f"\nTask Name {task_name}:\n"

                    if stderr:
                        logs += self.nomad_mgr.get_job_stderr(allocation.ID, task_name)
                    else:
                        logs += self.nomad_mgr.get_job_stdout(allocation.ID, task_name)

        return messages, logs.splitlines()  # type: ignore[reportReturnType]


def _get_parser() -> argparse.ArgumentParser:  # pragma: no-cover
    """
    Generate documentation; used by Sphinx.

    :meta private:
    """
    return NomadExecutor._get_parser()



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

import re
import warnings
import inspect
from collections.abc import Collection, Mapping, Sequence
from typing import Any, Callable, ClassVar

from airflow.sdk import Context
from airflow.sdk.bases.decorator import DecoratedOperator, TaskDecorator, task_decorator_factory
from airflow.sdk.definitions._internal.types import SET_DURING_EXECUTION
from airflow.utils.context import context_merge
# from airflow.utils.operator_helpers import determine_kwargs

from airflow.providers.nomad.operators.python import NomadPythonTaskOperator


class _NomadPythonTaskDecoratedOperator(DecoratedOperator, NomadPythonTaskOperator):
    """
    Wraps a Python callable and uses the callable return value as the Docker 'args' to be executed.

    :param python_callable: A reference to an object that is callable.
    :param op_kwargs: A dictionary of keyword arguments that will get unpacked
        in your function (templated).
    :param op_args: A list of positional arguments that will get unpacked when
        calling your callable (templated).
    """

    template_fields: Sequence[str] = (  # type: ignore [reportIncompatibleVariableOverride]
        *DecoratedOperator.template_fields,
        *NomadPythonTaskOperator.template_fields,
    )
    template_fields_renderers: ClassVar[dict[str, str]] = {
        **DecoratedOperator.template_fields_renderers,
        **NomadPythonTaskOperator.template_fields_renderers,
    }

    custom_operator_name: str = "@task.nomad"
    overwrite_rtif_after_execution: bool = True

    def __init__(
        self,
        *,
        python_callable: Callable,
        op_args: Collection[Any] | None = None,
        op_kwargs: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if kwargs.pop("multiple_outputs", None):
            warnings.warn(
                f"`multiple_outputs=True` is not supported in {self.custom_operator_name} tasks. Ignoring.",
                UserWarning,
                stacklevel=3,
            )

        if args := kwargs.pop("args", []):
            warnings.warn(
                "Use 'args' for Nomad Decorator with caution. Note that the output of the decorated function "
                + "is passed concatenated to 'args'",
                UserWarning,
                stacklevel=3,
            )

        super().__init__(
            python_callable=python_callable,
            op_args=op_args,
            op_kwargs=op_kwargs,
            args=args,
            multiple_outputs=False,
            python_command=SET_DURING_EXECUTION,
            **kwargs,
        )

    def _remove_docstring_from_source(self, lines: list[str]):
        if self.python_callable.__doc__:
            firstline = lines[0].strip()
            quotes = r"('''|\"\"\")"
            pat = re.compile(quotes)
            if pat.match(firstline):
                lines.pop(0)
                while lines and not pat.search(lines[0]):
                    lines.pop(0)
                lines.pop(0)
            elif re.match(r"^['|\"]", firstline):
                lines.pop(0)

    def _remove_header(self, lines):
        while lines and not re.match(" *def ", lines[0]):
            lines.pop(0)
        if lines:
            lines.pop(0)

    def _untab_lines(self, header: str, body_lines: list[str]) -> list[str]:
        spaces = ""
        if m := re.match(r"(?P<spaces>\s*)", header):
            spaces = m.groupdict().get("spaces", "")

        rmstr = spaces + " " * 4
        return [line.removeprefix(rmstr) for line in body_lines]

    def get_callable_source(self, fn: Callable) -> str | None:
        if not (source := inspect.getsource(fn)):
            self.log.error("No Python executable was found")
            return  # type: ignore[return-value]

        if not (lines := source.split("\n")):
            self.log.error(f"No Python executable {source} body was found")
            return  # type: ignore[return-value]

        if len(lines) == 1:
            lines = source.split(";")
            if not len(lines) > 1:
                self.log.error(
                    f"Python executable {lines} seems to consist of no more but a header"
                )
                return  # type: ignore[return-value]

        if not (head := lines[0]):
            self.log.error("Function definition seems wrong, header is empty")
            return  # type: ignore[return-value]

        self._remove_header(lines)
        self._remove_docstring_from_source(lines)
        lines = self._untab_lines(head, lines)

        return "\n".join(lines)

    def execute(self, context: Context) -> Any:
        context_merge(context, self.op_kwargs)
        # kwargs = determine_kwargs(self.python_callable, self.op_args, context)

        if not (cmd := self.get_callable_source(self.python_callable)):
            self.log.error("Nothing to execute")
            return  # type: ignore[return-value]

        self.python_command = cmd

        context["ti"].render_templates()  # type: ignore[attr-defined]

        return super().execute(context)


def nomad_python_task(
    python_callable: Callable | None = None,
    **kwargs,
) -> TaskDecorator:
    """
    Wrap a function into a NomadPythonTaskOperator.

    Accepts kwargs for operator kwargs. Can be reused in a single DAG. This function is only used only used
    during type checking or auto-completion.

    :param python_callable: Function to decorate.

    :meta private:
    """
    return task_decorator_factory(
        python_callable=python_callable,
        decorated_operator_class=_NomadPythonTaskDecoratedOperator,
        **kwargs,
    )
